@@ -48,10 +48,25 @@ use vars qw(@ISA $VERSION $DEBUG $SSL_ERROR $GLOBAL_CONTEXT_ARGS @EXPORT );
 	@EXPORT = qw( SSL_WANT_READ SSL_WANT_WRITE $SSL_ERROR GEN_DNS GEN_IPADD );
 }
 
+my @caller_force_inet4; # in case inet4 gets forced we store here who forced it
+my $can_ipv6;           # true if we successfully enabled ipv6 while loading
+
 BEGIN {
 	# Declare @ISA, $VERSION, $GLOBAL_CONTEXT_ARGS
-	@ISA = qw(IO::Socket::INET);
-	$VERSION = '1.15';
+	
+	# if we have IO::Socket::INET6 we will use this not IO::Socket::INET, because
+	# it can handle both IPv4 and IPv6. If we don't have INET6 available fall back
+	# to INET
+	if ( ! eval {
+		require Socket6;
+		Socket6->import( 'inet_pton' );
+		require IO::Socket::INET6;
+		@ISA = qw(IO::Socket::INET6);
+		$can_ipv6 = 1;
+	}) {
+		@ISA = qw(IO::Socket::INET);
+	}
+	$VERSION = '1.17';
 	$GLOBAL_CONTEXT_ARGS = {};
 
 	#Make $DEBUG another name for $Net::SSLeay::trace
@@ -113,13 +128,22 @@ sub import {
 	my @export;
 	foreach (@_) { 
 		if ( /^inet4$/i ) {
-			require IO::Socket::INET;
-			@ISA = 'IO::Socket::INET'
+			# explicitly fall back to inet4
+			@ISA = 'IO::Socket::INET';
+			@caller_force_inet4 = caller(); # save for warnings for 'inet6' case
 		} elsif ( /^inet6$/i ) {
-			require IO::Socket::INET6;
-			require Socket6;
-			Socket6->import( 'inet_pton' );
-			@ISA = 'IO::Socket::INET6'
+			# check if we have already ipv6 as base
+			if ( ! UNIVERSAL::isa( $class, 'IO::Socket::INET6' )) {
+				# either we don't support it or we disabled it by explicitly
+				# loading it with 'inet4'. In this case re-enable but warn
+				# because this is probably an error
+				if ( $can_ipv6 ) {
+					@ISA = 'IO::Socket::INET6';
+					warn "IPv6 support re-enabled in __PACKAGE__, got disabled in file $caller_force_inet4[1] line $caller_force_inet4[2]";
+				} else {
+					die "INET6 is not supported, missing Socket6 or IO::Socket::INET6";
+				}
+			}
 		} elsif ( /^:?debug(\d+)/ ) {
 			$DEBUG=$1;
 		} else {
@@ -141,6 +165,21 @@ sub configure {
 	my ($self, $arg_hash) = @_;
 	return _invalid_object() unless($self);
 
+	# work around Bug in IO::Socket::INET6 where it doesn't use the
+	# right family for the socket on BSD systems:
+	# http://rt.cpan.org/Ticket/Display.html?id=39550
+	if ( $can_ipv6 && ! $arg_hash->{Domain} && 
+		! ( $arg_hash->{LocalAddr} || $arg_hash->{LocalHost} ) &&
+		(my $peer = $arg_hash->{PeerAddr} || $arg_hash->{PeerHost})) {
+		# set Domain to AF_INET/AF_INET6 if there is only one choice
+		($peer, my $port) = IO::Socket::INET6::_sock_info( $peer,$arg_hash->{PeerPort},6 );
+		my @res = Socket6::getaddrinfo( $peer,$port,AF_UNSPEC,SOCK_STREAM );
+		if (@res == 5) {
+			$arg_hash->{Domain} = $res[0];
+			DEBUG(2,'set domain to '.$res[0] );
+		}
+	}
+
 	# force initial blocking 
 	# otherwise IO::Socket::SSL->new might return undef if the
 	# socket is nonblocking and it fails to connect immediatly
@@ -150,6 +189,7 @@ sub configure {
 
 	# because Net::HTTPS simple redefines blocking() to {} (e.g
 	# return undef) and IO::Socket::INET does not like this we
+		
 	# set Blocking only explicitly if it was set
 	$arg_hash->{Blocking} = 1 if defined ($blocking);
 
@@ -890,11 +930,11 @@ sub new_from_fd {
 		(my $mode = $_[0]) =~ tr/+<>//d;
 		shift unless length($mode);
 	}
-	my $handle = IO::Socket::INET->new_from_fd($fd, '+<')
+	my $handle = $ISA[0]->new_from_fd($fd, '+<')
 		|| return($class->error("Could not create socket from file descriptor."));
 
 	# Annoying workaround for Perl 5.6.1 and below:
-	$handle = IO::Socket::INET->new_from_fd($handle, '+<');
+	$handle = $ISA[0]->new_from_fd($handle, '+<');
 
 	return $class->start_SSL($handle, @_);
 }
@@ -1178,7 +1218,7 @@ sub set_ctx_defaults {
 
 sub opened {
 	my $self = shift;
-	return IO::Handle::opened($self) && ( ${*$self}{'_SSL_opened'} == 1 );
+	return IO::Handle::opened($self) && ${*$self}{'_SSL_opened'};
 }
 
 sub opening {
@@ -1308,7 +1348,7 @@ sub new {
 
 	if ($arg_hash->{'SSL_check_crl'}) {
 		if (Net::SSLeay::OPENSSL_VERSION_NUMBER() >= 0x0090702f) {
-			Net::SSLeay::X509_STORE_CTX_set_flags(
+			Net::SSLeay::X509_STORE_set_flags(
 				Net::SSLeay::CTX_get_cert_store($ctx),
 				Net::SSLeay::X509_V_FLAG_CRL_CHECK()
 			);
@@ -1882,6 +1922,12 @@ is expecting to read from or write to the socket and wants to be satisfied befor
 get to do anything. But with version 0.98 you are better comparing the global exported 
 variable $SSL_ERROR against the exported symbols SSL_WANT_READ and SSL_WANT_WRITE.
 
+=item B<opened()>
+
+This returns false if the socket could not be opened, 1 if the socket could be opened
+and the SSL handshake was successful done and -1 if the underlying IO::Handle is open,
+but the SSL handshake failed.
+
 =item B<< IO::Socket::SSL->start_SSL($socket, ... ) >>
 
 This will convert a glob reference or a socket that you provide to an IO::Socket::SSL
@@ -1984,6 +2030,19 @@ and read/sysread families instead.
 
 =back
 
+=head1 IPv6
+
+Support for IPv6 with IO::Socket::SSL is expected to work and basic testing is done.
+If IO::Socket::INET6 is available it will automatically use it instead of 
+IO::Socket::INET4. 
+
+Please be aware of the associated problems: If you give a name as a host and the
+host resolves to both IPv6 and IPv4 it will try IPv6 first and if there is no IPv6
+connectivity it will fail. 
+
+To avoid these problems you can either force IPv4 by specifying and AF_INET as the
+Domain (this is per socket) or load IO::Socket::SSL with the option 'inet4'
+(This is a global setting, e.g. affects all IO::Socket::SSL objects in the program).
 
 =head1 RETURN VALUES
 
@@ -1996,21 +2055,6 @@ The moral of the story: I<always> check the return values of these functions bef
 using them in any way that you consider meaningful.
 
 
-=head1 IPv6
-
-Support for IPv6 with IO::Socket::SSL is expected to work, but is experimental, as
-none of the author's machines use IPv6 and hence he cannot test IO::Socket::SSL with
-them.  However, a few brave people have used it without incident, so if you wish to
-make IO::Socket::SSL IPv6 aware, pass the 'inet6' option to IO::Socket::SSL when
-calling it (i.e. C<use IO::Socket::SSL qw(inet6);>).  You will need IO::Socket::INET6
-and Socket6 to use this option, and you will also need to write C<use Socket6;> before
-using IO::Socket::SSL.	If you absolutely do not want to use this (or want a quick
-change back to IPv4), pass the 'inet4' option instead.
-
-Currently, there is no support for using IPv4 and IPv6 simultaneously in a single program, 
-but it is planned for a future release.
-
-
 =head1 DEBUGGING
 
 If you are having problems using IO::Socket::SSL despite the fact that can recite backwards
@@ -2056,6 +2100,10 @@ SSL_password_cb.
 IO::Socket::SSL does not work together with Storable::fd_retrieve/fd_store.
 See BUGS file for more information and how to work around the problem.
 
+Non-blocking and timeouts (which are based on non-blocking) are not
+supported on Win32, because the underlying IO::Socket::INET does not support
+non-blocking on this platform.
+
 =head1 LIMITATIONS
 
 IO::Socket::SSL uses Net::SSLeay as the shiny interface to OpenSSL, which is
